unit AlfaBetaU;
{
 chi usa questa unit deve implementare i metodi astratti di
 TAbstractPosition e TAbstractMove.
 P.e.:
  type
    TMove = class(TAbstractMove)
    private
      ... implementazione interna ...
    public
      ... implementazione metodi astratti ...
    end;

    TPosition = class(TAbstractPosition)
    private
      ... implementazione interna ...
    public
      ... implementazione metodi astratti ...
    end;
}

interface

uses SysUtils;

type
  { scelta dell'algoritmo }
  TAlgorithm = (alMinimax,alAlfabeta);

  { liste di ritorno della getAllMoves }
  TMovesList = array of string;
  TValuesList = array of integer;

  { TAbstractDebug }
  TDebugMode = (dmNormal,dmExtra);
  TAbstractDebug = class
    procedure Open; virtual; abstract;
    procedure Write(DM: TDebugMode; FormatString: string;
                     Args: array of const); virtual; abstract;
    procedure Close; virtual; abstract;
  end;

  { TAbstractMove }
  TAbstractPosition = class; // definita subito dopo
  TAbstractMove = class
    // valore speciale che indica non definito
    // (usato nel caso ci sia gi una vittoria nella config. data)
    procedure setUndefined; virtual; abstract;
    // stabilisce se la mossa  definita
    function isDefined: boolean; virtual; abstract;
    // riempie la move con la prima mossa legale valida nella posizione
    // se non ci sono mosse legali deve impostare la mossa a undefined
    procedure First(Position: TAbstractPosition); virtual; abstract;
    // riempie la move con la prossima mossa legale nella posizione
    // se le mosse legali sono finite ritorna false e non modifica l'oggetto
    function Next(Position: TAbstractPosition): boolean; virtual; abstract;
    // copia una mossa su un'altra
    procedure CopyFrom(SrcMove: TAbstractMove); virtual; abstract;
    // ritorna una rappresentazione della mossa in formato stringa
    // (per scopi di debug)
    function AsString: string; virtual; abstract;
  end;

  { TAbstractPosiion }
  TAbstractPosition = class
    // stabilisce se la posizione  legale
    function isLegal: boolean; virtual; abstract;
    // istanzia uma TMove vuota
    function CreateMove: TAbstractMove; virtual; abstract;
    // fa la mossa sulla schacchiera
    procedure DoMove(Mossa: TAbstractMove); virtual; abstract;
    // toglie la mossa dalla scacchiera
    procedure UndoMove(Mossa: TAbstractMove); virtual; abstract;
    // parte della funzione di valutazione che controlla se  un nodo
    // terminale (vittoria, sconfitta o patta)
    // Se ritorna false verr chiamata anche Euristics
    function isLeaf(var Value: integer): boolean;
      virtual; abstract;
    // tutto quello che resta fuori della funzione di valutazione
    // (Value viene ritornato)
    function Euristics: integer;
      virtual; abstract;
    // controlla se Value  un valore di vittoria
    class function hasWon(Value: integer): boolean; virtual; abstract;
    // ritorna una rappresentazione della posizione in formato stringa
    // (per scopi di debug)
    function AsString: string; virtual; abstract;
  end;

{ entry-point 1 - ritorna la mossa migliore }
function ChooseMove(Algorithm: TAlgorithm; Debug: TAbstractDebug;
  Depth: integer; Position: TAbstractPosition;
  var Move: string; var Value: integer): boolean;
{ entry-point 2 - ritorna tutte le mosse }
function getAllMoves(Algorithm: TAlgorithm; Debug: TAbstractDebug;
  Depth: integer; Position: TAbstractPosition;
  var Moves: TMovesList; var Values: TValuesList): boolean;

implementation

{ Algoritmo minimax senza potatura alfa-beta }
procedure Minimax(Depth: integer; Position: TAbstractPosition;
  BestMove: TAbstractMove; var BestValue: integer);
    forward;
{ Algoritmo minimax con potatura alfa-beta }
procedure AlfaBeta(Depth: integer; Position: TAbstractPosition;
  Alfa,Beta: integer; BestMove: TAbstractMove; var BestValue: integer);
    forward;

{ puntatore globale alla variabile di debug }
var
  Dbg: TAbstractDebug;

{ funzione di infinito che ritorna il numero pi basso possibile}
// il valore ritornato  invertibile, p.e. menoinfinito := -infinity
// (in tal caso -infinito NON  il numero intero pi piccolo)
function infinity: integer;
begin
  // funziona perch  sempre High(integer) = |Low(integer)| - 1
  // result := High(integer)
  result := 1000;
end;

{ entry-point 1 - ritorna la mossa migliore }
function ChooseMove(Algorithm: TAlgorithm; Debug: TAbstractDebug;
  Depth: integer; Position: TAbstractPosition;
  var Move: string; var Value: integer): boolean;
var
  BestMove: TAbstractMove;
begin
  if Position.isLegal then
  begin
    Debug.Open;
    try
      Dbg := Debug; // copia nella variabile globale
      BestMove := Position.CreateMove;
      try
        if Algorithm = alAlfabeta then
          // alfa = -infinito, beta = +infinito
          AlfaBeta(Depth,Position,-infinity,infinity,BestMove,Value)
        else
          Minimax(Depth,Position,BestMove,Value);
        Move := BestMove.AsString;
        Value := -Value;
      finally
        BestMove.Free
      end;
      result := true
    finally
      Debug.Close;
    end
  end
  else result := false
end;

{ entry-point 2 - ritorna tutte le mosse }
function getAllMoves(Algorithm: TAlgorithm; Debug: TAbstractDebug;
  Depth: integer; Position: TAbstractPosition;
  var Moves: TMovesList; var Values: TValuesList): boolean;
var
  CurrentMove: TAbstractMove;
  Cnt: integer;
  Value: integer;
begin
  Depth := Depth - 1;
  if (Depth >= 0) and Position.isLegal then
  begin
    Debug.Open;
    try
      Dbg := Debug; // copia nella variabile globale
      Moves := nil; Values := nil;
      Cnt := 0;
      CurrentMove := Position.CreateMove;
      try
        CurrentMove.First(Position);
        if CurrentMove.isDefined then
          repeat
            Position.DoMove(CurrentMove);
            if Algorithm = alAlfabeta then
              // alfa = -infinito, beta = +infinito
              AlfaBeta(Depth,Position,-infinity,infinity,CurrentMove,Value)
            else
              Minimax(Depth,Position,CurrentMove,Value);
            SetLength(Moves,Cnt+1);
            SetLength(Values,Cnt+1);
            Moves[Cnt] := CurrentMove.AsString;
            Values[Cnt] := -Value;
            Inc(Cnt);
            Position.UndoMove(CurrentMove);
          until not CurrentMove.Next(Position);
      finally
        CurrentMove.Free
      end;
      result := true
    finally
      Debug.Close;
    end
  end
  else result := false
end;

{ Algoritmo minimax senza potatura alfa-beta }
procedure Minimax(Depth: integer; Position: TAbstractPosition;
  BestMove: TAbstractMove; var BestValue: integer);
var
  { mossa dummy }
  dummyMove: TAbstractMove;
  { mossa e valore correnti }
  CurrentMove: TAbstractMove;
  CurrentValue: integer;
  { solo per scopi di debug }
  OldMoveStr: string;
  OldValue: integer;
begin
  if Position.isLeaf(CurrentValue) then
  begin
    BestMove.setUndefined;
    BestValue := CurrentValue;
    Dbg.Write(dmNormal,'VITTORIA %d: %s = %d',
               [Depth,Position.AsString,CurrentValue])
  end
  else if (Depth = 0) then
  begin
    BestMove.setUndefined;
    CurrentValue := Position.Euristics;
    BestValue := CurrentValue;
    Dbg.Write(dmNormal,'Foglia   0: %s = %d',
              [Position.AsString,CurrentValue])
  end
  else
  begin
    Dbg.Write(dmNormal,'Interno  %d: %s',[Depth,Position.AsString]);
    Dec(Depth);
    { Evaluate and choose }
    BestMove.setUndefined;
    BestValue := -infinity;
    CurrentMove := Position.CreateMove;
    try
      CurrentMove.First(Position);
      if not CurrentMove.isDefined then
        raise Exception.Create('Errore: posizione non terminale non ha mosse legali');
      repeat
        Position.DoMove(CurrentMove);
        Dbg.Write(dmExtra,'%d: ora entro in minimax per la mossa %s',
                  [Depth,CurrentMove.AsString]);
        dummyMove := Position.CreateMove;
        try
          Minimax(Depth,Position,dummyMove,CurrentValue);
        finally
          dummyMove.Free;
        end;
        Position.UndoMove(CurrentMove);
        Dbg.Write(dmExtra,'%d: minimax ha ritornato %d per la mossa %s',
                   [Depth,CurrentValue,CurrentMove.AsString]);
        { update - massimizza sempre }
        OldMoveStr := BestMove.AsString;
        OldValue := BestValue;
        if CurrentValue > BestValue then
        begin
          BestMove.CopyFrom(CurrentMove);
          BestValue := CurrentValue
        end;
        Dbg.Write(dmExtra,'valutazione (%s,%d) --> (%s,%d)',
                   [OldMoveStr,OldValue,BestMove.AsString,BestValue]);
        { cut-off }
        if Position.hasWon(CurrentValue) then
        begin
          // gi fatto da update
          // BestMove.CopyFrom(CurrentMove);
          // BestValue := CurrentValue
          Dbg.Write(dmExtra,'taglio vittoria (%s,%d)',
                   [BestMove.AsString,BestValue]);
          break
        end
      until not CurrentMove.Next(Position);
    finally
      CurrentMove.Free;
    end;
    { clausola finale di Evaluate and choose + chiusura minimax }
    BestValue := -BestValue;
    Dbg.Write(dmNormal,'Chiudi   %d  - sceglie (%s,%d)',
              [Depth,BestMove.AsString,-BestValue]);
  end
end;

{ algoritmo minimax con potatura alfa-beta }
procedure AlfaBeta(Depth: integer; Position: TAbstractPosition;
  Alfa,Beta: integer;
  BestMove: TAbstractMove; var BestValue: integer);
var
  Temp: integer; // variabile temporanea
  { mossa dummy }
  dummyMove: TAbstractMove;
  { mossa corrente }
  CurrentMove: TAbstractMove;
  CurrentValue: integer;
begin
  if Position.isLeaf(CurrentValue) then
  begin
    BestMove.setUndefined;
    BestValue := CurrentValue;
    Dbg.Write(dmNormal,'VITTORIA %d: %s = %d',
               [Depth,Position.AsString,CurrentValue])
  end
  else if (Depth = 0) then
  begin
    BestMove.setUndefined;
    CurrentValue := Position.Euristics;
    BestValue := CurrentValue;
    Dbg.Write(dmNormal,'Foglia   0: %s = %d --> %d,%d',
              [Position.AsString,CurrentValue,Alfa,Beta])
  end
  else
  begin
    Dbg.Write(dmNormal,'Interno  %d: %s     --> %d,%d',
               [Depth,Position.AsString,Alfa,Beta]);
    // alfa = -beta <-> beta = -alfa
    Temp := Alfa; Alfa := -Beta; Beta := -Temp;
    Dec(Depth);
    { Evaluate and choose }
    BestMove.setUndefined;
//    BestValue := -infinity;
    CurrentMove := Position.CreateMove;
    try
      CurrentMove.First(Position);
      if not CurrentMove.isDefined then
        raise Exception.Create('Errore: posizione non terminale non ha mosse legali');
      repeat
        Position.DoMove(CurrentMove);
        Dbg.Write(dmExtra,'%d: ora entro in alfabeta per la mossa %s',
                  [Depth,CurrentMove.AsString]);
        dummyMove := Position.CreateMove;
        try
          AlfaBeta(Depth,Position,Alfa,Beta,dummyMove,CurrentValue);
        finally
          dummyMove.Free;
        end;
        Position.UndoMove(CurrentMove);
        Dbg.Write(dmExtra,'%d: alfabeta ha ritornato %d per la mossa %s',
                  [Depth,CurrentValue,CurrentMove.AsString]);
        if (CurrentValue > Alfa) then
        begin
          if (CurrentValue >= Beta) then
            Dbg.Write(dmExtra,'%d: valutazione: %d>=%d --> Best := %s (stop)',
                       [Depth,CurrentValue,Beta,CurrentMove.AsString])
          else
            Dbg.Write(dmExtra,
              '%d: valutazione: %d<%d<%d --> alfa := %d; Best := %s (continua)',
              [Depth,Alfa,CurrentValue,Beta,CurrentValue,CurrentMove.AsString]);
          BestMove.CopyFrom(CurrentMove);
          Alfa := CurrentValue;
          if (CurrentValue >= Beta) then break
        end
        else
          Dbg.Write(dmExtra,'%d: valutazione: %d<=%d --> continua',
                    [Depth,CurrentValue,Alfa])
      until not CurrentMove.Next(Position);
    finally
      CurrentMove.Free;
    end;
    { clausola finale di Evaluate and choose }
//    BestMove := Move;
    BestValue := -Alfa;
    Dbg.Write(dmNormal,'Chiudi   %d  - sceglie (%s,%d)',
              [Depth,BestMove.AsString,Alfa]);
  end
end;

end.
