unit RegoleU;
(* In questo file ci sono solo le regole del gioco richiamate
dal motore (minimax o alfabeta) *)

interface

uses SysUtils,AlfaBetaU;

type
  TMove = class(TAbstractMove)
  private
    FChoice: integer;
    function cercaMossaDa(Position: TAbstractPosition; idx: integer): integer;
  public
    property Choice: integer read FChoice;
    // valore speciale che indica non definito
    // (usato nel caso ci sia gi una vittoria nella config. data)
    procedure setUndefined; override;
    // stabilisce se la mossa  definita
    function isDefined: boolean; override;
    // copia una mossa su un'altra
    procedure CopyFrom(SrcMove: TAbstractMove); override;
    // ritorna la prima mossa legale
    procedure First(Position: TAbstractPosition); override;
    // riempie la TMove passata con la prossima mossa legale
    // se le mosse legali sono finite ritorna false e non modifica l'oggetto
    function Next(Position: TAbstractPosition): boolean; override;
    // ritorna una rappresentazione della mossa in formato stringa
    // (per scopi di debug)
    function AsString: string; override;
  end;

  TChessboard = array[1..9] of char;
  TPosition = class(TAbstractPosition)
  private
    FChb: TChessboard;
    Pezzo: char;
    function CalcolaPezzo: char;
  public
    // accesso pubblico alla scacchiera
    property Chessboard: TChessboard read FChb write FChb;
    // imposta la scacchiera con una stringa
    procedure setChessboard(b: string);
    // ritorna il numero di mosse mancanti
    function RemainingMoves: integer;
    { Interfaccia con l'algoritmo }
    // stabilisce se la scacchiera  legale
    function isLegal: boolean; override;
    // istanzia uma TMove vuota
    function CreateMove: TAbstractMove; override;
    // fa la mossa sulla schacchiera
    procedure DoMove(Mossa: TAbstractMove); override;
    // toglie la mossa dalla scacchiera
    procedure UndoMove(Mossa: TAbstractMove); override;
    // parte della funzione di valutazione che controlla se  un nodo
    // terminale (vittoria, sconfitta o patta)
    // Se ritorna false verr chiamata anche Euristics
    function isLeaf(var Value: integer): boolean;
      override;
    // tutto quello che resta fuori della funzione di valutazione
    // (Value viene ritornato)
    function Euristics: integer;
      override;
    // controlla se Value  un valore di vittoria
    class function hasWon(Value: integer): boolean; override;
    // ritorna una rappresentazione della posizione in formato stringa
    // (per scopi di debug)
    function AsString: string; override;
  end;

implementation

{ TMove }

procedure TMove.setUndefined;
begin
  FChoice := 0 // se questo valore cambia ci saranno problemi
end;

// 2) isLegalMove(Position,Move)
// verifica se Move  una mossa lecita da Position
// ATTENZIONE!! NON VIENE USATO!!
procedure TMove.CopyFrom(SrcMove: TAbstractMove);
begin
  FChoice := TMove(SrcMove).FChoice
end;

function TMove.AsString: string;
begin
  result := IntToStr(FChoice)
end;

procedure TMove.First(Position: TAbstractPosition);
begin
  FChoice := cercaMossaDa(Position,Low(TPosition(Position).FChb)); // 0 se non ce ne sono
end;

function TMove.Next(Position: TAbstractPosition): boolean;
var
  idx: integer;
begin
  idx := cercaMossaDa(Position,FChoice+1);
  if idx > 0 then
  begin
    FChoice := idx;
    result := true
  end
  else result := false
end;

{ genera le mosse }
function TMove.cercaMossaDa(Position: TAbstractPosition; idx: integer): integer;
var
  i: integer;
  Pos: TPosition;
begin
  Pos := TPosition(Position);
  result := 0;
  for i := idx to High(Pos.FChb) do
    if Pos.FChb[i] = ' ' then
    begin
      result := i;
      break
    end
end;

function TMove.isDefined: boolean;
begin
  result := FChoice <> 0
end;

{ TPosition }

function TPosition.CreateMove: TAbstractMove;
begin
  result := TMove.Create
end;

// isLegal
// stabilisce se la scacchiera  legale
function TPosition.isLegal: boolean;
var
  i: integer;
begin
  result := true;
  for i := Low(FChb) to High(FChb) do
    if not (FChb[i] in [' ','X','O']) then
    begin
      result := false;
      break
    end;
  if Pezzo <> CalcolaPezzo then result := false
end;

// 3) DoMove(Ply,Mossa,VecchiaScacchiera,NuovaScacchiera)
//    esegue una mossa (controllo di legalit gi eseguito)
//    (data la mossa aggiorna la posizione)
procedure TPosition.DoMove(Mossa: TAbstractMove);
begin
  if Pezzo = 'X' then Pezzo := 'O'
  else Pezzo := 'X';
  FChb[TMove(Mossa).FChoice] := Pezzo;
end;

procedure TPosition.UndoMove(Mossa: TAbstractMove);
begin
  if Pezzo = 'X' then Pezzo := 'O'
  else Pezzo := 'X';
  FChb[TMove(Mossa).FChoice] := ' '
end;

(*
 Valuate(Ply, Flag: integer; var Value: integer): boolean
 funzione di valutazione
 Input: ply, flag minimax (1/-1)
 Output: value
 Result: se true  una foglia
*)
function TPosition.isLeaf(var Value: integer): boolean;
    function Vince(P: char): boolean;
    begin
      result :=
      ((P = FChb[1]) and (P = FChb[2]) and (P = FChb[3])) or
      ((P = FChb[4]) and (P = FChb[5]) and (P = FChb[6])) or
      ((P = FChb[7]) and (P = FChb[8]) and (P = FChb[9])) or
      ((P = FChb[1]) and (P = FChb[4]) and (P = FChb[7])) or
      ((P = FChb[2]) and (P = FChb[5]) and (P = FChb[8])) or
      ((P = FChb[3]) and (P = FChb[6]) and (P = FChb[9])) or
      ((P = FChb[1]) and (P = FChb[5]) and (P = FChb[9])) or
      ((P = FChb[3]) and (P = FChb[5]) and (P = FChb[7]));
    end;
begin
  if Vince(Pezzo) then
  begin
    Value := 1;
    result := true
  end
  else
    result := false
end;

function TPosition.Euristics: integer;
begin
  // nessuna euristica
  result := 0
end;

procedure TPosition.setChessboard(b: string);
var
  i: integer;
begin
  if Length(b) = 9 then
    for i := 1 to 9 do FChb[i] := b[i]
  else raise Exception.Create('cannot initialize position: wrong string');
  Pezzo := CalcolaPezzo;
end;

// 6) Won(Value)
// Ha successo se  un valore di vittoria
// NOTA: Viene usato solo da minimax (NON da alfabeta)
// NB2: in futuro anche alfabeta potrebbe usarlo
class function TPosition.hasWon(Value: integer): boolean;
begin
  result := Value = 1
end;

function TPosition.CalcolaPezzo: char;
var
  Ply: integer;
begin
  Ply := 9 - RemainingMoves;
  if (Ply mod 2) = 1 then result := 'X'
  else result := 'O';
end;

function TPosition.RemainingMoves: integer;
var
  i: integer;
begin
  result := 0;
  for i := Low(FChb) to High(FChb) do
    if FChb[i] = ' ' then Inc(Result)
end;

function TPosition.AsString: string;
begin
  result := Pezzo + ', ' + FChb
end;

end.
