/*
AlfaBeta
Algoritmo minimax con potatura alfa-beta
Modificato dal minimax di pag. 288 di "Art Of Prolog" (Sterling/Shapiro)

Abbiamo 2 entry-point:
1 - per avere solo la mossa migliore usare:
    choose_move(alfabeta,+Depth,+Position,-Move,-Value[,+Mode,+DebugFile]]) 
2 - per avere tutta la rosa di mosse con i valori associati usare:
    get_all_moves(alfabeta,+Depth,+Position,-List[,+Mode,+DebugFile]]) 
Argomenti:
- Depth  il valore di profondit dell'algoritmo
- Position  l'input (tipicamente la scacchiera + info sul giocatore)
- Mode=normal/debug/extra  un parametro opzionale per il debug
- DebugFile  il nome del file di debug (default: 'debug.txt')
- Move,Value  la mossa migliore col valore associato
- List  una lista di coppie (Mossa,Valore)

Funzioni da implementare esternamente:
1) legal_position(+Position)
   valuta se l'input  legale
2) move(+Position,-Move) 
   genera una alla volta tutte le mosse legali
3) move(+Move,+OldPosition,-NewPosition) 
   data la mossa aggiorna la posizione
4) is_leaf(+Position, -Value) 
   funzione di valutazione per condizione di fine partita
5) euristics(+Position, -Value)
   funzione di valutazione per le situazioni restanti
6) won(+Value)
   dice se il valore della f. di valutazione  di vittoria
------------------------------------------------------------------
*/

:- [debug].
:- multifile choose_move/5, choose_move/7.
:- multifile get_all_moves/4, get_all_moves/6.

% get_all_moves
% ritorna la lista di tutte le mosse legali
% NB: per il momento ritorna solo la mossa migliore
get_all_moves(alfabeta,Depth,Position,List) :-
  get_all_moves(alfabeta,Depth,Position,List,none,nil).
get_all_moves(alfabeta,Depth,Position,List,Mode,DebugFile) :-
  D1 is Depth-1, D1 >= 0, !,
  open_debug(Mode,DebugFile,Dbg),  
  legal_position(Position),
  findall(M,move(Position,M),Moves),
  evaluate_ab(Position,D1,Moves,List,Dbg),
  close_debug(Dbg).
evaluate_ab(_,_,[],[],_) :- !.
evaluate_ab(OldPos,Depth,[Move|Moves],[(Move,V)|Others],Dbg) :-
  move(Move,OldPos,NewPos),
  alfa_beta(Depth,NewPos,-1000,1000,_,V,Dbg), !,
  evaluate_ab(OldPos,Depth,Moves,Others,Dbg).

/*
choose_move(alfabeta,+Depth,+Position,-Move,-Value[,+Mode,+DebugFile]])
sceglie la mossa migliore
*/
choose_move(alfabeta,Depth,Position,Move,Value) :-
  choose_move(alfabeta,Depth,Position,Move,Value,none,nil).
choose_move(alfabeta,Depth,Position,Move,Value,Mode,DebugFile) :-
  open_debug(Mode,DebugFile,Dbg),  
  legal_position(Position),
  alfa_beta(Depth,Position,-1000,1000,Move,V,Dbg),
  Value is -V,
  close_debug(Dbg).

alfa_beta(Depth,Position,_,_,_,Value,Dbg):- 
  is_leaf(Position,Value), !,
%  Value1 is Value*MaxMin,
  string_position(Position,S), 
  writeDebug(Dbg,'VITTORIA %w: %w = %w',[Depth,S,Value]).
alfa_beta(0,Position,Alfa,Beta,_,Value,Dbg):- 
  euristics(Position,Value), !,
%  Value1 is Value*MaxMin,
  string_position(Position,S),
  writeDebug(Dbg,'Foglia   0: %w = %w --> %w,%w',[S,Value,Alfa,Beta]).
alfa_beta(D,Position,Alfa,Beta,Move,Value1,Dbg) :-
%  D > 0,
  bagof(M,move(Position,M),Moves),
  D1 is D-1,
  Alfa1 is -Beta, Beta1 is -Alfa,
  string_position(Position,S),
  writeDebug(Dbg,'Interno  %w: %w     --> %w,%w',[D,S,Alfa,Beta]),
  evaluate_choose_ab(Moves,Position,D1,Alfa1,Beta1,nil,(Move,Value),Dbg),
  Value1 is -Value.

/*
evaluate_choose_ab(+Moves,+Position,+Depth,
                   +Alfa,+Beta,+Record,-BestMove,+Mode,+Dbg) 
valuta e scegli
*/
evaluate_choose_ab([Move|Moves],Position,Depth,Alfa,Beta,Move1,BestMove,Dbg) :-
  move(Move,Position,NewPos),
  string_move(Move,SM),
  extraDebug(Dbg,'%w: ora entro in alfabeta per la mossa %w',[Depth,SM]),
  alfa_beta(Depth,NewPos,Alfa,Beta,_,Value,Dbg),
  extraDebug(Dbg,'%w: alfabeta ha ritornato %w per la mossa %w',[Depth,Value,SM]), 
  cutoff_ab(Move,Value,Depth,Alfa,Beta,Moves,Position,Move1,BestMove,Dbg), !.
evaluate_choose_ab([],_,Depth,Alfa,_,Move,(Move,Alfa),Dbg) :-
  !, string_move(Move,MStr),
  writeDebug(Dbg,'Chiudi   %w  - sceglie (%w,%w)',[Depth,MStr,Alfa]).
cutoff_ab(Move,Value,Depth,_,Beta,_,_,_,(Move,Value),Dbg) :- 
  Value>=Beta, !, string_move(Move,MStr),
  extraDebug(Dbg,'%w: valutazione: %w>=%w --> Best := %w (stop)',
             [Depth,Value,Beta,MStr]).
cutoff_ab(Move,Value,D,Alfa,Beta,Moves,Position,_,BestMove,Dbg) :-
  Alfa<Value, Value<Beta, !,
  string_move(Move,MStr),
  extraDebug(Dbg,'%w: valutazione: %w<%w<%w --> alfa := %w; Best := %w (continua)',
             [D,Alfa,Value,Beta,Value,MStr]),
  evaluate_choose_ab(Moves,Position,D,Value,Beta,Move,BestMove,Dbg).
cutoff_ab(_,Value,D,Alfa,Beta,Moves,Position,Move1,BestMove,Dbg) :-
  Value=<Alfa, !,
  extraDebug(Dbg,'%w: valutazione: %w<=%w --> continua',[D,Value,Alfa]),
  evaluate_choose_ab(Moves,Position,D,Alfa,Beta,Move1,BestMove,Dbg).









