/*
minimax
Algoritmo minimax (senza potatura alfa-beta)
Modificato dal minimax di pag. 288 di "Art Of Prolog" (Sterling/Shapiro)
Correzioni:
1 - per un errore di stampa la chiamata di minimax in evaluate_choose passava Move
    invece di variabile muta (MoveX)
2 - aggiunta riga finale in evaluate_choose che inverte il risultato di Value in fase
    di backtracking
3 - aggiunto taglio in evaluate_choose senza il quale si ha presto un debordamento dello
    stack
4 - aggiunto caso base per minimax per i nodi terminali
Solo minimax: 
5 - aggiunto predicato cut_off che taglia il controllo dei fratelli in
    caso di raggiunta vittoria

Abbiamo 2 entry-point:
1 - per avere solo la mossa migliore usare:
    choose_move(minimax,+Depth,+Position,-Move,-Value[,+Mode,+DebugFile]]) 
2 - per avere tutta la rosa di mosse con i valori associati usare:
    get_all_moves(minimax,+Depth,+Position,-List[,+Mode,+DebugFile]]) 
Argomenti:
- Depth  il valore di profondit dell'algoritmo
- Position  l'input (tipicamente la scacchiera + info sul giocatore)
- Mode=normal/debug/extra  un parametro opzionale per il debug
- DebugFile  il nome del file di debug (default: 'debug.txt')
- Move,Value  la mossa migliore col valore associato
- List  una lista di coppie (Mossa,Valore)

Funzioni da implementare esternamente:
1) legal_position(+Position)
   valuta se l'input  legale
2) move(+Position,-Move) 
   genera una alla volta tutte le mosse legali
3) move(+Move,+OldPosition,-NewPosition) 
   data la mossa aggiorna la posizione
4) is_leaf(+Position, -Value) 
   funzione di valutazione per condizione di fine partita
5) euristics(+Position, -Value)
   funzione di valutazione per le situazioni restanti
6) won(+Value)
   dice se il valore della f. di valutazione  di vittoria
7) string_position(+Position,-String)
   ritorna la position in formato stringa per il debug
8) string_move(+Move,-String)
   ritorna la move in formato stringa per il debug
------------------------------------------------------------------
*/

:- [debug].
:- multifile choose_move/5, choose_move/7.
:- multifile get_all_moves/4, get_all_moves/6.

% get_all_moves
% ritorna la lista di tutte le mosse legali
get_all_moves(Algoritmo,Depth,Position,List) :-
  get_all_moves(Algoritmo,Depth,Position,List,none,nil).
get_all_moves(Algoritmo,Depth,Position,List,Mode,DebugFile) :-
  is_minimax(Algoritmo), !,
  D1 is Depth-1, D1 >= 0, !,
  open_debug(Mode,DebugFile,Dbg),  
  legal_position(Position),
  findall(M,move(Position,M),Moves),
  evaluate(Algoritmo,Position,D1,Moves,List,Dbg),
  close_debug(Dbg).
evaluate(_,_,_,[],[],_) :- !.
evaluate(Algoritmo,OldPos,Depth,[Move|Moves],[(Move,V)|Others],Dbg) :-
  move(Move,OldPos,NewPos),
  minimax(Algoritmo,Depth,NewPos,_,V,Dbg), !,
  evaluate(Algoritmo,OldPos,Depth,Moves,Others,Dbg).

/*
choose_move(minimax,+Depth,+Position,-Move,-Value[,+Mode,+DebugFile]])
sceglie la mossa migliore
*/
choose_move(Algoritmo,Depth,Position,Move,Value) :-
  choose_move(Algoritmo,Depth,Position,Move,Value,none,nil).
choose_move(Algoritmo,Depth,Position,Move,Value,Mode,DebugFile) :-
  is_minimax(Algoritmo), !,
  open_debug(Mode,DebugFile,Dbg),  
  legal_position(Position),
  minimax(Algoritmo,Depth,Position,Move,V,Dbg),
  Value is -V,
  close_debug(Dbg).

is_minimax(minimax).
is_minimax(minimax_simple).

minimax(_,Depth,Position,_,V,Dbg) :-
  is_leaf(Position,V), !,
  string_position(Position,S), 
  writeDebug(Dbg,'VITTORIA %w: %w = %w',[Depth,S,V]).
minimax(_,0,Position,_,V,Dbg) :- 
  euristics(Position,V), !,
  string_position(Position,S),
  writeDebug(Dbg,'Foglia   0: %w = %w',[S,V]).
minimax(Algoritmo,D,Position,NewMove,Value,Dbg) :-
%  D > 0,
  bagof(M,move(Position,M),Moves),
  D1 is D - 1,
  string_position(Position,S),
  writeDebug(Dbg,'Interno  %w: %w',[D,S]),
  evaluate_choose(Algoritmo,Moves,Position,D1,(nil,-1000),(NewMove,V),Dbg),
  Value is -V.

/*
evaluate_choose(+Moves,+Position,+Depth,+Flag,+Record,-BestMove,+Dbg) 
valuta e scegli
Sceglie la migliore mossa Best=(Move,Value) dall'insieme delle mosse Moves
dalla posizione attuale Position usando l'algoritmo Minimax
spingendosi fino al livello Depth. 
Flag (1 o -1) indica se al momento stiamo minimizzando o massimizzando. 
Record registra l'attuale miglior mossa
*/
evaluate_choose(Algoritmo,[Move|Moves],Position,Depth,Record,Best,Dbg) :-
  move(Move,Position,NewPos),
  string_move(Move,SM),
  extraDebug(Dbg,'%w: ora entro in minimax per la mossa %w',[Depth,SM]),
  minimax(Algoritmo,Depth,NewPos,_,Value,Dbg),
  extraDebug(Dbg,'%w: minimax ha ritornato %w per la mossa %w',
            [Depth,Value,SM]), 
  cut_off(Algoritmo,Move,Moves,Value,Position,Depth,Record,Best,Dbg), !.
evaluate_choose(_,[],_,Depth,(M,V),(M,V),Dbg) :-
  !, string_move(M,MStr),
  writeDebug(Dbg,'Chiudi   %w  - sceglie (%w,%w)',[Depth,MStr,V]).
cut_off(minimax,Move,_,Value,_,_,_,Best,Dbg) :-
  won(Value), !, Best = (Move,Value),
  extraDebug(Dbg,'taglio vittoria (%w)',[Best]).
cut_off(Algoritmo,Move,Moves,Value,Position,Depth,Record,Best,Dbg) :-
  update(Move,Value,Record,NewRecord),
  extraDebug(Dbg,'valutazione %w --> %w',[Record,NewRecord]), 
  evaluate_choose(Algoritmo,Moves,Position,Depth,NewRecord,Best,Dbg), !.

% update(+M1,+V1,(+M2,+V2),(-M,-V))
% sceglie la mossa con valore massimo
update(_,Value,(Move1,Value1),(Move1,Value1)) :-
  Value =< Value1.
update(Move,Value,(_,Value1),(Move,Value)) :-
  Value > Value1.

/*
% value(+Position, -Value)
value(Position,Value) :-
  is_leaf(Position,Value) ;
  euristics(Position,Value).
*/


















