/*
Filetto
In questo file ci sono le regole del gioco richiamate 
dal motore (minimax o alfabeta)
In particolare ci sono le seguenti funzioni:
1) legal_position(+Position)
   valuta se l'input  legale
2) move(+Position,-Move) 
   genera una alla volta tutte le mosse legali
3) move(+Move,+OldPosition,-NewPosition) 
   data la mossa aggiorna la posizione
4) is_leaf(+Position, -Value) 
   funzione di valutazione per condizione di fine partita
5) euristics(+Position, -Value)
   funzione di valutazione per le situazioni restanti
6) won(+Value)
   dice se il valore della f. di valutazione  di vittoria
*/

%%%%%%%%%%%%%%%%
% ENTRY POINTS %
%%%%%%%%%%%%%%%%

computer_choice(Algoritmo,MaxDepth,Scacchiera,List) :-
  computer_choice(Algoritmo,MaxDepth,Scacchiera,List,none).
computer_choice(Algoritmo,MaxDepth,Scacchiera,List,Mode) :-
  computer_choice(Algoritmo,MaxDepth,Scacchiera,List,Mode,'debug.txt').
computer_choice(Algoritmo,MaxDepth,Scacchiera,List,Mode,DebugFile) :-
  calcola_depth(Scacchiera,MaxDepth,D),
  pezzo_iniziale(Scacchiera,Pezzo),
  get_all_moves(Algoritmo,D,(Pezzo,Scacchiera),List,Mode,DebugFile).
computer_move(Algoritmo,MaxDepth,Scacchiera,Move,Value) :-
  computer_move(Algoritmo,MaxDepth,Scacchiera,Move,Value,none).
computer_move(Algoritmo,MaxDepth,Scacchiera,Move,Value,Mode) :-
  computer_move(Algoritmo,MaxDepth,Scacchiera,Move,Value,Mode,'debug.txt').
computer_move(Algoritmo,MaxDepth,Scacchiera,Move,Value,Mode,DebugFile) :-
  calcola_depth(Scacchiera,MaxDepth,D),
  pezzo_iniziale(Scacchiera,Pezzo),
  choose_move(Algoritmo,D,(Pezzo,Scacchiera),Move,Value,Mode,DebugFile).

%%%%%%%%%%
% REGOLE %
%%%%%%%%%%

% legal_position(+Position)
% -------------------------
% stabilisce se la posizione  legale
% Position = (Pezzo,Scacchiera)
% Pezzo = 'X'|'O'
% Scacchiera  una lista di 9 elementi 'X'|'O'|' '
legal_position((Pezzo,Scacchiera)) :-
  pezzo_valido(Pezzo),
  length(Scacchiera,9), caselle_lecite(Scacchiera), !.
legal_position(_) :- write('Scacchiera illegale'),nl,fail.
caselle_lecite([]).
caselle_lecite([T|C]) :- casella_lecita(T),caselle_lecite(C).
casella_lecita(C) :-
  char_code('X',C), !; % X
  char_code('O',C), !; % O
  char_code(' ',C), !. % casella vuota
pezzo_valido('X').
pezzo_valido('O').

% move(+Position,?Move)
% ---------------------
% verifica se Move  una mossa lecita da Position
% oppure
% genera la prossima mossa legale
% RICHIAMATA ANCHE DALL'INTERFACCIA
move((_,Scacchiera),Move) :-
  char_code(' ',Spazio), nth1(Move,Scacchiera,Spazio).

% move(+Move,+OldPosition,-NewPosition)
% -------------------------------------
% data la mossa aggiorna la posizione
% RICHIAMATA ANCHE DALL'INTERFACCIA
move(Move,(OldPezzo,OldS),(NewPezzo,NewS)) :-
  cambia_pezzo(OldPezzo,NewPezzo),
  separa_lista(Move,OldS,L1,E,L2),
  char_code(' ',Spazio), Spazio=E,
  char_code(NewPezzo,P),append(L1,[P|L2],NewS).
% separa_lista(+Idx, +SourceList, -List1, -Elem, -List2)
% spezza la SourceList in 3 pezzi (lista+elem.+lista) all'indice Idx
separa_lista(Idx,List,L1,E,L2) :-
  Idx1 is Idx-1, taglia_lista(Idx1,List,L1,[E|L2]).
% taglia_lista(+Idx, +SourceList, -List1, -List2)
taglia_lista(0,L,[],L) :- !.
taglia_lista(Idx,[T|C],[T|L1],L2) :-
  /*Idx > 0,*/ Idx1 is Idx-1, !, taglia_lista(Idx1,C,L1,L2).
taglia_lista(_,[],[],[]).

% is_leaf(+Position, -Value)
% --------------------------
% funzione di valutazione
% Input: pezzo, scacchiera
% Data una posizione dice se c' una vittoria del giocatore corrente e 
% nel caso ritorna il valore della funzione di valutazione, altrimenti
% fallisce
% RICHIAMATA ANCHE DALL'INTERFACCIA
is_leaf((Pezzo,Scacchiera),Val) :-
  char_code(Pezzo,P),vince_pezzo(P,Scacchiera),!,Val is 1.

% vince_pezzo(+Pezzo,+Scacchiera) :-
% controlla se c' una vittoria del pezzo
vince_pezzo(Pezzo,[C1,C2,C3,C4,C5,C6,C7,C8,C9]) :-
  fa_filetto(Pezzo,[C1,C2,C3]);
  fa_filetto(Pezzo,[C4,C5,C6]);
  fa_filetto(Pezzo,[C7,C8,C9]);
  fa_filetto(Pezzo,[C1,C4,C7]);
  fa_filetto(Pezzo,[C2,C5,C8]);
  fa_filetto(Pezzo,[C3,C6,C9]);
  fa_filetto(Pezzo,[C1,C5,C9]);
  fa_filetto(Pezzo,[C3,C5,C7]).
fa_filetto(Pezzo,[Pezzo,Pezzo,Pezzo]).

% euristics(+Position, -Value)
% ----------------------------
% funzione di valutazione
% Input: pezzo, scacchiera
euristics(_,0). 

% won(+Value)
% -----------
% Ha successo se  un valore di vittoria
won(1).

% string_position(+Position,-String)
string_position((Pezzo,Scacchiera),String) :-
  name(S,Scacchiera), String =(Pezzo,S).

% string_move(+Move,-String)
string_move(Move,Move).

%%%%%%%%%
% UTILS %
%%%%%%%%%

% calcola_depth(+Scacchiera,+MaxDepth,-Depth)
calcola_depth(Scacchiera,MaxDepth,Depth) :-
  remaining_moves(Scacchiera,RestDepth),
  Depth is min(MaxDepth,RestDepth).

% pezzo_iniziale(+Scacchiera,-Pezzo)
pezzo_iniziale(Scacchiera,Pezzo) :-
  remaining_moves(Scacchiera,RestDepth),
  Ply is 9 - RestDepth,
  sceglie_pezzo(Ply,Pezzo).

% usato dall'algoritmo
cambia_pezzo('X','O').
cambia_pezzo('O','X').

% sceglie_pezzo(+Ply,-Pezzo)
% Usato dal chiamante per fornire il pezzo iniziale iniziale
sceglie_pezzo(Ply,'X') :- dispari(Ply),!.

sceglie_pezzo(_,'O').
dispari(N) :- 1 =:= N mod 2.

% remaining_moves(+Scacchiera,-RemainingMoves) 
% Usato dal chiamante per calcolare Depth
remaining_moves([],0).
remaining_moves([T|S],Rest) :- 
  char_code(' ',Spazio), T \= Spazio, !, remaining_moves(S,Rest).
remaining_moves([T|S],Rest) :- 
  char_code(' ',Spazio), T = Spazio, remaining_moves(S,Rest1),
  Rest is Rest1+1.

% piena(+Scacchiera)
% controlla se non ci sono pi mosse disponibili
% RICHIAMATA SOLO DALL'INTERFACCIA
piena(Scacchiera) :- char_code(' ',Spazio), \+ memberchk(Spazio,Scacchiera).
/*
% vecchia implementazione:
patta([],patta).
patta([Casella|Coda) :- char_code(C,Casella),C \= ' ',patta(Coda).
*/










